__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

"""
execute the command line tool
"""

from .cl_tool import execute_cl_tool

execute_cl_tool()
